use std::fmt;

#[derive(Debug, Clone, PartialEq)]
pub struct StatusError;

impl fmt::Display for StatusError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{self:?}")
    }
}

impl std::error::Error for StatusError {}

#[derive(Debug, Clone, PartialEq)]
#[repr(u8)]
pub enum Status {
    /// These codes are reserved, and must not be sent by a Misfin server.
    Input = 10,
    /// Status codes beginning with 2 are SUCCESS status codes, which mean
    /// that the client's message has been delivered.
    Success = 20,
    /// The mailbox has moved to a different address, and this message should
    /// be resent to that address.
    Redirect(Redirect) = 30,
    /// Status codes beginning with 5 are PERMANENT FAILURE status codes, which
    /// mean the request did not succeed, and should not be retried.
    TemporaryFailure(TemporaryFailure) = 40,
    /// Something is wrong with the mailserver, and you should not try to
    /// resend your message.
    PermanentFailure(PermanentFailure) = 50,
    /// Status codes beginning with 6 are AUTHENTICATION FAILURE status codes,
    /// which mean that there was an issue with the client's certificate.
    AuthenticationFailure(AuthenticationFailure) = 60,
}

impl TryFrom<u8> for Status {
    type Error = StatusError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            n if n / 10 == 1 => Ok(Self::Input),
            n if n / 20 == 1 => Ok(Self::Success),
            n if n / 30 == 1 => Ok(Self::Redirect((value % 30).try_into()?)),
            n if n / 40 == 1 => Ok(Self::TemporaryFailure((value % 40).try_into()?)),
            n if n / 50 == 1 => Ok(Self::PermanentFailure((value % 50).try_into()?)),
            n if n / 60 == 1 => Ok(Self::AuthenticationFailure((value % 60).try_into()?)),
            _ => Err(StatusError),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum Redirect {
    /// The mailbox has moved to a different address, and this message
    /// should be resent to that address.
    Temporary = 0,
    /// The mailbox has moved to a different address, and all future
    /// messages should be sent to that address.
    Permanent = 1,
    Other,
}

impl TryFrom<u8> for Redirect {
    type Error = StatusError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::Temporary),
            1 => Ok(Self::Permanent),
            n if n < 10 => Ok(Self::Other),
            _ => Err(StatusError),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum TemporaryFailure {
    /// The mailserver experienced a transient issue, and the message
    /// should be resent.
    TemporaryError = 0,
    /// The mailserver can't accept mail right now.
    ServerUnavailable = 1,
    /// A mailserver script ran for your message, but experienced an error.
    CgiError = 2,
    /// There was a problem accepting mail for that domain, but it might
    /// resolve itself.
    ProxyError = 3,
    /// You are being rate limited - wait before trying to send more mail.
    RateLimit = 4,
    /// The mailbox isn't accepting mail right now, but it might in the future.
    MailboxFull = 5,
    Other,
}

impl TryFrom<u8> for TemporaryFailure {
    type Error = StatusError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::TemporaryError),
            1 => Ok(Self::ServerUnavailable),
            2 => Ok(Self::CgiError),
            3 => Ok(Self::ProxyError),
            4 => Ok(Self::RateLimit),
            5 => Ok(Self::MailboxFull),
            n if n < 10 => Ok(Self::Other),
            _ => Err(StatusError),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum PermanentFailure {
    PermanentError = 0,
    MailboxNonexistent = 1,
    MailboxGone = 2,
    DomainNotServiced = 3,
    BadRequest = 9,
    Other,
}

impl TryFrom<u8> for PermanentFailure {
    type Error = StatusError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::PermanentError),
            1 => Ok(Self::MailboxNonexistent),
            2 => Ok(Self::MailboxGone),
            3 => Ok(Self::DomainNotServiced),
            9 => Ok(Self::BadRequest),
            n if n < 10 => Ok(Self::Other),
            _ => Err(StatusError),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum AuthenticationFailure {
    CertificateRequired = 0,
    UnauthorizedSender = 1,
    CertificateInvalid = 2,
    IdentityMismatch = 3,
    ProofRequired = 4,
    Other,
}

impl TryFrom<u8> for AuthenticationFailure {
    type Error = StatusError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::CertificateRequired),
            1 => Ok(Self::UnauthorizedSender),
            2 => Ok(Self::CertificateInvalid),
            3 => Ok(Self::IdentityMismatch),
            4 => Ok(Self::ProofRequired),
            n if n < 10 => Ok(Self::Other),
            _ => Err(StatusError),
        }
    }
}
